# RDSR
ENHANCED_STRUCTURED_REPORT_UID = "1.2.840.10008.5.1.4.1.1.88.22"
XRAY_RADIATION_DOSE_STRUCTURED_REPORT_UID = "1.2.840.10008.5.1.4.1.1.88.67"
RADIATION_DOSE_STRUCTURED_REPORT_SOP_CLASS_UIDS = [
    ENHANCED_STRUCTURED_REPORT_UID,
    XRAY_RADIATION_DOSE_STRUCTURED_REPORT_UID,
]

# Secondary Capture
SECONDARY_CAPTURE_UID = "1.2.840.10008.5.1.4.1.1.7"
SECONDARY_CAPTURE_SOP_CLASS_UIDS = [SECONDARY_CAPTURE_UID]
